@@ -75,10 +75,10 @@ GLOBAL_STR(gStr5, R"zZXx(Builtin Commands
   [Child Process] jobs        wait
                   fg        X bg       X kill        X disown 
   [External]      test [      getopts
-  [Conditional]   cmd/true    cmd/false               colon :
+  [Conditional]   cmd/true    cmd/false                colon :
   [Introspection] help        hash       cmd/type    X caller
   [Word Lookup]   command     builtin
-  [Interactive]   alias       unalias    history     X fc     X bind
+  [Interactive]   alias       unalias    history       fc     X bind
 X [Unsupported]   enable
 )zZXx");
 
@@ -108,8 +108,8 @@ Example:
 
 GLOBAL_STR(gStr7, R"zZXx(Command Language
 
-  [Commands]      simple-command               prefix-binding
-                  semicolon ;    
+  [Commands]      simple-command               command-lookup-order
+                  prefix-binding               semicolon ;    
   [Conditional]   case           if            dbracket [[
                   bang !         and &&        or ||
   [Iteration]     while          until         for            for-expr-sh ((
@@ -162,6 +162,7 @@ GLOBAL_STR(gStr10, R"zZXx(Global Shell Options
   [Interactive]    emacs           vi
   [Compat]         eval_unsafe_arith            ignore_flags_not_impl
                    ignore_shopt_not_impl
+  [Optimize]       rewrite_extern
 )zZXx");
 
 GLOBAL_STR(gStr11, R"zZXx(Assignments and Expressions
@@ -241,6 +242,7 @@ Examples:
 
 GLOBAL_STR(gStr17, R"zZXx(Word Language
 
+  [Joining]       osh-word-join 'single'\'$myvar"double $x"
   [Quotes]        osh-string    'abc'  $'line\n'  "$var"
   [Substitutions] command-sub   $(date)   `date`
                   var-sub       ${var}   $0   $9   
@@ -305,9 +307,9 @@ GLOBAL_STR(gStr19, R"zZXx(Builtin Commands
                   shvar                  Temporary modify global settings
                   ctx                    Share and update a temporary "context"
                   push-registers         Save registers like $?, PIPESTATUS
-  [Introspection] runproc                Run a proc; use as main entry point
+  [Word Lookup]   invoke                 Run a command, and control name lookup
+                  runproc                Run a proc; use as main entry point
                 X extern                 Run an external command, with an ENV
-                X invoke                 Control which "invokables" are run
   [Modules]       
                   source-guard           guard against duplicate 'source'
                   is-main                false when sourcing a file
@@ -318,7 +320,9 @@ GLOBAL_STR(gStr19, R"zZXx(Builtin Commands
                   ysh-wait               wait --all --verbose
                   write                  Like echo, with --, --sep, --end
                   fork         forkwait  Replace & and (), and takes a block
-                  fopen                  Open multiple streams, takes a block
+                  redir                  Run a block, with redirects
+  [Private]       cat          rm        POSIX-compatible
+                  sleep
   [Hay Config]    hay          haynode   For DSLs and config files
   [Completion]    compadjust   compexport
   [Data Formats]  json                   read write
@@ -328,24 +332,23 @@ GLOBAL_STR(gStr19, R"zZXx(Builtin Commands
 GLOBAL_STR(gStr20, R"zZXx(Builtin Functions
 
   [Values]        len()             func/type()
-  [Conversions]   bool()            int()            float()
-                  str()             list()           dict()
+  [Conversions]   bool()            int()              float()
+                  str()             list()             dict()
                 X runes()         X encodeRunes()
                 X bytes()         X encodeBytes()
   [Str]         X strcmp()          shSplit()
   [List]          join()       
-  [Dict]          keys()            values()         get()       
-  [Float]         floatsEqual()   X isinf()        X isnan()
-  [Obj]           first()           rest()           get()
+  [Dict]          keys()            values()           get()       
+  [Float]         floatsEqual()   X isinf()          X isnan()
+  [Obj]           first()           rest()             get()
   [Word]          glob()            maybe()
   [Serialize]     toJson()          fromJson()
                   toJson8()         fromJson8()
                 X toJ8Line()      X fromJ8Line()
-X [Quoting]       quoteJ8()         quoteSh()        quoteHtml()
-  [Pattern]       _group()          _start()         _end()
-  [Reflection]    func/eval()       func/evalExpr()
-  [Introspect]    shvarGet()        getVar()         setVar()  
-                  parseCommand()  X parseExpr()    X bindFrame()
+  [Pattern]       _group()          _start()           _end()
+  [Reflection]    func/eval()       func/evalExpr()  
+  [Introspect]    shvarGet()        getVar()           setVar()  
+                  parseCommand()  X parseExpr()      X bindFrame()
   [Hay Config]    parseHay()        evalHay()
 X [Hashing]       sha1dc()          sha256()
 )zZXx");
@@ -379,6 +382,7 @@ Example:
 GLOBAL_STR(gStr22, R"zZXx(Command Language
 
   [Commands]      simple-command
+                  command-lookup-order
                   ysh-prefix-binding
                   semicolon ;
   [Redirects]     ysh-here-str    read <<< '''
@@ -458,6 +462,7 @@ GLOBAL_STR(gStr25, R"zZXx(Other Mini Languages
 
 GLOBAL_STR(gStr26, R"zZXx(Global Shell Options
 
+  [Optimize]     rewrite_extern
   [Groups]       strict:all      ysh:upgrade     ysh:all
   [YSH Details]  opts-redefine   opts-internal
 )zZXx");
@@ -470,7 +475,7 @@ GLOBAL_STR(gStr27, R"zZXx(Plugins and Hooks
 GLOBAL_STR(gStr28, R"zZXx(Special Variables
 
   [YSH Vars]      ARGV                ENV
-                  __defaults__        __builtins__
+                  __defaults__        __builtins__        X __sh_function__
                   _this_dir
   [YSH Status]    _error
                   _pipeline_status    _process_sub_status
@@ -483,7 +488,7 @@ GLOBAL_STR(gStr28, R"zZXx(Special Variables
                   OILS_GC_STATS       OILS_GC_STATS_FD
                   LIB_YSH
   [Float]         NAN                 INFINITY
-  [Module]        __provide__
+  [Module]        __provide__ 
   [Other Env]     HOME                PATH
 )zZXx");
 
@@ -523,8 +528,11 @@ GLOBAL_STR(gStr30, R"zZXx(Types and Methods
                    Dict        erase()        X Dict/clear()    X accum()
                              X update()
                    Place       setValue()
-  [Code Types]     Func        BuiltinFunc      BoundFunc
-                   Proc        BuiltinProc
+  [Code Types]     Func        
+                   BuiltinFunc
+                   BoundFunc
+                   Proc        docComment()   
+                   BuiltinProc
   [Objects]        Obj         __invoke__       new
                              X __call__         __index__       X __str__
   [Reflection]     Command     sourceCode()
@@ -555,6 +563,7 @@ Note that bin/ysh is the same as bin/osh with the ysh:all option group set:
 
 GLOBAL_STR(gStr32, R"zZXx(Word Language
 
+  [Joining]       ysh-word-join    --flag='value'  PATH="value"
   [Quotes]        ysh-string       "x is $x"  $"x is $x"  r'[a-z]\n'
                                    u'line\n'  b'byte \yff'
                   triple-quoted    """  $"""  r'''  u'''  b'''
@@ -563,7 +572,7 @@ GLOBAL_STR(gStr32, R"zZXx(Word Language
                   expr-splice      echo @[split(x)]
                   var-splice       @myarray @ARGV
                   command-sub      $(date)
-                  ysh-command-sub  @(cat my-j8-lines.txt)
+                  command-splice   @(cat my-j8-lines.txt)
   [Formatting]  X ysh-printf       ${x %.3f}
                 X ysh-format       ${x|html}
 )zZXx");
@@ -1177,9 +1186,19 @@ func parseArgs(spec, argv) {
 
   var value
   var found
+  var escape_remaining = false
   while (i < argc) {
     var arg = argv[i]
-    if (arg.startsWith('-')) {
+
+    if (escape_remaining) {
+      call rest->append(arg)
+    } elif (arg === '--') {
+      if (not spec.rest) {
+        error "Unexpected '--' argument - extraneous positional arguments are prohibited" (code=2)
+      }
+
+      setvar escape_remaining = true
+    } elif (arg.startsWith('-')) {
       setvar found = false
 
       for flag in (spec.flags) {
@@ -1457,7 +1476,212 @@ func sum(list; start=0) {
 }
 )zZXx");
 
-GLOBAL_STR(gStr45, R"zZXx(# stream.ysh
+GLOBAL_STR(gStr45, R"zZXx(const __provide__ = :| sh shell ninja make |
+
+# Issues for quoting:
+#
+# 1. What is the alphabet we're quoting to?
+#    - Output can be "all bytes", "all unicode strings", or "ASCII"
+# 2. Start with a simple algorithm to quote everything
+#    - POSIX shell may take ' to ''\'''
+# 3. Heuristic that may avoid quotes, to make it more readable
+#    - But really it should be \'
+#   - If the ' appears at the beginning or the end, we could have a different
+#     algorithm.  Or we could strip leading and trailing ''
+# 4. Are there any byte strings / unicode strings that can't be quoted?
+#    - e.g. NUL bytes?
+#    - for JSON/JS, binary strings?  Only Unicode strings can be encoded.
+#
+# Builtins:
+#   toJson() toJson8() 
+#   toJ8Line() - does the "maybe unquoted" logic
+#
+# Related functions:
+#   encode.base{16,32,64} - Crockford has a base32
+#   decode.base{16,32,64}
+#   Also base85, and maybe base58 base36
+#
+# In Python: bin() oct() hex() int(i, 9) and %o %x (there is no %b)
+#
+# Other:
+#   Punycode for Unicode domain names uses xn-- ?
+#   CSS has escapes with \
+#   HTTP Cookies have "" and \?
+#
+# Related:
+#   demo/url-search-params.ysh is the PARSER / unquoter for quote.urlParam()
+
+func sh(s) {
+  ### Quote POSIX sh string
+
+  # replace ' with sequence ' \' '
+  # Note: the \\ is hard to read - '' doesn't help
+  return ("'" ++ s.replace("'", "'\\''") ++ "'")
+}
+
+func shell(s) {
+  ### Quote shell string, where bash and zsh style $'\n' is allowed
+
+  # TODO: Binding for j8_lite.MaybeShellEncode / ShellEncode
+  return (toJson(s))
+}
+
+func ysh(s) {
+  ### Quote YSH string as b'\n'
+
+  # TODO: Binding for j8_lite.YshEncode(unquoted_ok)
+  return (toJson(s))
+}
+
+#
+# Build Tools
+#
+
+func make(s) {
+  var out = []
+  var n = len(s)
+  for i in (0 ..< n) {
+    var ch = s[i]
+
+    case (ch) {
+      (\r) |
+      (\n) {
+        error "Can't quote newlines for Make?"
+      }
+      ('$') {
+        call out->append('$')
+        call out->append('$')
+      }
+      ('%') |
+      (r'\') |
+      # glob characters
+      ('[') | (']') | ('*') | ('?') {
+        call out->append(r'\')
+        call out->append(ch)
+      }
+      (else) {
+        call out->append(ch)
+      }
+    }
+  }
+  return (join(out, ''))
+}
+
+# https://ninja-build.org/manual.html#ref_lexer
+# $ escapes newline, space, : and $
+# and then the rest is interpreted by /bin/sh
+func ninja(s) {
+  var out = []
+  var n = len(s)
+  for i in (0 ..< n) {
+    var ch = s[i]
+
+    case (ch) {
+      # Subtlety: Ninja allows $ to escape a newline, but it's only for
+      # line continuations - for breaking long lists of files.
+      # - A file itself should not have a newline.
+      # - It strips literal newlines out of shell commands.
+      # So disallow it
+      (\r) |
+      (\n) {
+        error "Can't quote newlines for Ninja"
+      }
+      ('$') |
+      (' ') |
+      #('#') |  # Ninja has no way to escape comments!
+      (':') {
+        call out->append('$')
+        call out->append(ch)
+      }
+      (else) {
+        call out->append(ch)
+      }
+    }
+  }
+  return (join(out, ''))
+}
+
+# I don't know the rule here - uses \?
+func cmake(s) {
+  return (s)
+}
+
+#
+# Tables, Objects, Documents
+#
+
+func csv(s) {
+  # double up " etc.
+  return (s)
+}
+
+func sql(s) {
+  # double up ' etc.
+  return (s)
+}
+
+# quote.json is just toJson()
+
+func html(s) {
+  ### Quote shell string, where bash and zsh style $'\n' is allowed
+
+  # Binding for j8_lite.MaybeShellEncode / ShellEncode
+  return (toJson(s))
+}
+
+#
+# Web
+#
+
+func urlParam(s) {
+  # urllib.quote
+  # 'foo bar %' -> 'foo+bar %AB'
+  return (toJson(s))
+}
+
+#
+# Programming Languages
+#
+
+# Python 2 or 3
+func py(s) {
+  return (s)
+}
+
+# C or C++
+# Can it be Unicode?
+func c(s) {
+  return (s)
+}
+
+# quote.js is just toJson() ?
+# But it can't handle binary strings?
+# We can make a table?
+
+#
+# Windows
+#
+
+# We want something that works everywhere -- it should never DOUBLE-ESCAPE, but
+# it can "unnecessarily" ESCAPE certain characters.
+
+# C runtime
+func win32_crt(s) {
+  return (s)
+}
+
+# win32 cmd.exe batch files?
+func win32_cmd(s) {
+  return (s)
+}
+
+# batch different than cmd.exe?
+func win32_batch(s) {
+  return (s)
+}
+)zZXx");
+
+GLOBAL_STR(gStr46, R"zZXx(# stream.ysh
 #
 # Usage:
 #   source --builtin stream.ysh
@@ -1769,7 +1993,7 @@ run-byo-tests() {
 byo-maybe-run
 )zZXx");
 
-GLOBAL_STR(gStr46, R"zZXx(# table.ysh - Library for tables.
+GLOBAL_STR(gStr47, R"zZXx(# table.ysh - Library for tables.
 #
 # Usage:
 #   source --builtin table.ysh
@@ -1921,7 +2145,7 @@ proc summary {
 byo-maybe-run
 )zZXx");
 
-GLOBAL_STR(gStr47, R"zZXx(#!/usr/bin/env bash
+GLOBAL_STR(gStr48, R"zZXx(#!/usr/bin/env bash
 #
 # Testing library for bash and OSH.
 #
@@ -2024,9 +2248,10 @@ TextFile array[] = {
     {.rel_path = "stdlib/ysh/def.ysh", .contents = gStr42},
     {.rel_path = "stdlib/ysh/list.ysh", .contents = gStr43},
     {.rel_path = "stdlib/ysh/math.ysh", .contents = gStr44},
-    {.rel_path = "stdlib/ysh/stream.ysh", .contents = gStr45},
-    {.rel_path = "stdlib/ysh/table.ysh", .contents = gStr46},
-    {.rel_path = "stdlib/ysh/yblocks.ysh", .contents = gStr47},
+    {.rel_path = "stdlib/ysh/quote.ysh", .contents = gStr45},
+    {.rel_path = "stdlib/ysh/stream.ysh", .contents = gStr46},
+    {.rel_path = "stdlib/ysh/table.ysh", .contents = gStr47},
+    {.rel_path = "stdlib/ysh/yblocks.ysh", .contents = gStr48},
 
     {.rel_path = nullptr, .contents = nullptr},
 };
